/* imgmirc.dll is a plug-in for the mIRC application
* Copyright (C) 2003-2005 Ivan Zvolsky
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version. 
* 
* You should have received a copy of the GNU Lesser General Public
* License along with this library (lesser.txt)
*/
#include "stdafx.h"
#include "imgmirc.h"

#include "msgs_hook.h"
#include "kbd_hook.h"
#include "convthread.h"
#include "flags.h"
#include "my_funcs.h"
#include "pcre.h"
#include "list_manage.h"
#include "imgstorage.h"
#include "replstorage.h"
#include "locale.h"
#include "configure_dlg.h"

HANDLE hMod;

char mircData[900]="";
char mircDir [900];

HWND mircWnd;
DWORD idInst = 0;
HCONV hCmdConv;
HCONV hEvalConv;

HANDLE h_loaded;
bool succ_loaded = FALSE;

// needed to selectively intercept API
WORD mirc_minor_ver;
WORD mirc_major_ver;

typedef struct tag_find 
{
	char class_name[32];
	HWND hwnd;
} _find_str;

_find_str find;

BOOL CALLBACK FindStatic(HWND hwnd,LPARAM lParam);

BOOL APIENTRY DllMain( HANDLE hModule, 
					  DWORD  ul_reason_for_call, 
					  LPVOID lpReserved
					  )
{
	switch (ul_reason_for_call)
	{
	case DLL_PROCESS_ATTACH:
	case DLL_THREAD_ATTACH:
		hMod = hModule;	
		break;
	case DLL_THREAD_DETACH:
	case DLL_PROCESS_DETACH:
		break;
	}
	return TRUE;
}

int __stdcall UnloadDll(int mTimeout)
{
	if (mTimeout == 1 && _KEEP_ALIVE_)  // 1 means unload by timeout
		return 0;

	CloseHandle (h_loaded);
	CloseHandle (hDlgAllowed);
	CloseHandle (hDlgDestroyed);
	if (hhk_msgs) UnhookWindowsHookEx(hhk_msgs);
	if (hhk_kbd) UnhookWindowsHookEx(hhk_kbd);

	return_functions ();

	if (_PCRE_HIGHLIGHTS_)
		conv_command (hCmdConv, "/im_hldel -1"); // delete all patterns and free memory

	if (!_hl_item::ctables)
	{
		free ((void *)_hl_item::ctables);
		setlocale(LC_CTYPE, "C");
	}

	conv_command (hCmdConv,"/dll imgmirc.dll extern_disconnect EVALUATE");
	verb_conv_echo ("imgmirc: EVALUATE conversation ended.",7);

	if (_VERBOSE_) 
		conv_command (hCmdConv,"/dll imgmirc.dll extern_disconnect COMMAND | /echo 7 -a imgmirc: COMMAND conversation ended."); 
	else
		conv_command (hCmdConv,"/dll imgmirc.dll extern_disconnect COMMAND");

	DdeUninitialize (idInst);

	if (_PCRE_HIGHLIGHTS_ && hl_privmsg) delete hl_privmsg;

	std::vector<_repl_item *>::iterator i;
	for (i=repl_list.begin(); i<repl_list.end(); i++)
		delete (*i);

	return 1;
}

void __stdcall LoadDll(LOADINFO* loadinfo)
{
	mircWnd = loadinfo->mHwnd;

	h_loaded = CreateEvent (NULL, true, true, "imgmirc");
	hDlgDestroyed = CreateEvent (NULL, false, false, "dlgdestroyed");
	hDlgAllowed = CreateEvent (NULL, false, false, "dlgallowed");

	DdeInitialize (&idInst, (PFNCALLBACK)DdeCallback, APPCMD_CLIENTONLY, 0);

	hCmdConv = DdeConnect ( idInst, 
		DdeCreateStringHandle (idInst,"mIRC",CP_WINANSI),
		DdeCreateStringHandle (idInst,"COMMAND",CP_WINANSI),
		NULL);
	hEvalConv = DdeConnect (idInst, 
		DdeCreateStringHandle (idInst,"mIRC",CP_WINANSI),
		DdeCreateStringHandle (idInst,"EVALUATE",CP_WINANSI),
		NULL);

	mirc_major_ver = LOWORD(loadinfo->mVersion);
	mirc_minor_ver = HIWORD(loadinfo->mVersion);
}

int __stdcall set_kbd_hook(HWND mWnd, HWND aWnd, char *data, 
						   char *parms, BOOL show, BOOL nopause)
{
	if (!hhk_kbd)
	{
		DWORD thID = GetWindowThreadProcessId(mircWnd, &thID);
		hhk_kbd =  SetWindowsHookEx(WH_KEYBOARD,GetKbdProc,(HINSTANCE)hMod,thID);
	}	
	return 1;
}

int __stdcall set_quit_hook(HWND mWnd, HWND aWnd, char *data, 
							char *parms, BOOL show, BOOL nopause)
{
	if (!hhk_msgs)
	{
		DWORD thID = GetWindowThreadProcessId(mircWnd, &thID);
		hhk_msgs =  SetWindowsHookEx(WH_GETMESSAGE,GetMsgProc,(HINSTANCE)hMod,thID);
	}
	return 1;
}

int __stdcall extern_disconnect(HWND mWnd, HWND aWnd, char *data, 
								char *parms, BOOL show, BOOL nopause)
{
	if (strstr(data,"COMMAND")) {DdeDisconnect (hCmdConv);return 1;}
	if (strstr(data,"EVALUATE")) {DdeDisconnect (hEvalConv);return 1;}
	return 1;
}

int __stdcall extern_evaluate(HWND mWnd, HWND aWnd, char *data, 
							  char *parms, BOOL show, BOOL nopause)
{
	evaluate (data);
	conv_echo (3,"%s",mircData);
	return 1;
}

int __stdcall preload(HWND mWnd, HWND aWnd, char *data, 
					  char *parms, BOOL show, BOOL nopause)
{
	if (WaitForSingleObject (h_loaded ,0) == WAIT_TIMEOUT)
	{
		conv_echo (4, "imgmirc.dll already loaded.");
		return 1;
	}

	succ_loaded = TRUE;
	ResetEvent (h_loaded);

	conv_command ("//set %%imgmirc_ver 0.27s");

	load_all_flags();

	if (_AUTOSET_QHOOK_)
	{
		conv_command (hCmdConv,"/dll imgmirc.dll set_quit_hook");
		verb_conv_echo ("imgmirc: quit message hooked.",7);
	}

	if (_AUTOSET_KBDHOOK_)
	{
		conv_command (hCmdConv,"/dll imgmirc.dll set_kbd_hook");
		verb_conv_echo ("imgmirc: keyboard hooked.",7);
	}

	memset (mircData, 0, 900);
	evaluate ("$mircdir");
	sprintf (mircDir, "%s", mircData);

	try 
	{
		char * tmp;
		if (_PCRE_HIGHLIGHTS_)
		{
			// setting up a locale:
			if (!(tmp = setlocale(LC_ALL, "")))
				conv_echo ("imgmirc: setlocale failed.",4);
			else verb_conv_echo (tmp, 3);

			if (!(_hl_item::ctables = pcre_maketables()))
				conv_echo ("pcre: maketables failed.", 4);

			hl_list.reserve (32);
			hl_privmsg = new _hl_item ("\\:.*?PRIVMSG.*?\\x20\\:",false);
			conv_command (hCmdConv, "/im_hlload");

			conv_command ("//set %%hl_mainstr $+($chr($rand(128,255)),$chr($rand(128,255)),$chr($rand(128,255)))");
			evaluate ("%hl_mainstr");
			memcpy (hl_mainstr, mircData, 3);
		}
		if (_SMILES_) LoadReplacementsList();

		verb_conv_echo(7, "mirc version: %d.%d", mirc_major_ver, mirc_minor_ver);
		do_intercept ();
	}
	catch (int i)
	{
		switch (i)
		{
		case 8:conv_echo ("imgmirc: failed to open replacements.txt",7);break;
		default:conv_echo ("imgmirc: failed to precompile PRIVMSG prefix.dll",7);
		}
	}

	SetEvent (hDlgAllowed);

	verb_conv_echo ("\x2imgmirc\x2 dll started.",7);
	return 1;	
}

int __stdcall set(HWND mWnd, HWND aWnd, char *data, 
				  char *parms, BOOL show, BOOL nopause)
{
	int n, val;
	if (sscanf(data,"%d %d",&n,&val) != 2) return 1;

	flags[n].flag = val;
	return 1;
}

int __stdcall configure(HWND mWnd, HWND aWnd, char *data, 
						char *parms, BOOL show, BOOL nopause)
{
	if ((WaitForSingleObject (h_loaded ,0) == WAIT_OBJECT_0))
	{
		sprintf (data, "/dll imgmirc.dll preload");
		return 2;
	}

	if (!succ_loaded) 
	{
		conv_command (hCmdConv,"/.timerimgmirc_tryconfigure off");
		return 1;
	}

	if ((WaitForSingleObject (hDlgAllowed, 1) == WAIT_TIMEOUT)) return 1;

	ResetEvent (hDlgDestroyed);
	ResetEvent (hDlgAllowed);
	conv_command (hCmdConv,"/.timerimgmirc_changetrack -imo 0 100 /dll imgmirc.dll save_config");
	evaluate ("imgmirc v $+ %imgmirc_ver configurator");
	DWORD thID;
	hDlgThread = CreateThread (NULL, 0, DlgThreadProc, 0, 0, &thID);
	return 1;
}

int __stdcall save_config(HWND mWnd, HWND aWnd, char *data, 
						  char *parms, BOOL show, BOOL nopause)
{
	if (WaitForSingleObject (hDlgDestroyed, 1) == WAIT_TIMEOUT) 
		return 1;
	conv_command ("/.timerimgmirc_changetrack off");

	if (refresh) 
	{
		save_all_flags (TRUE);
		conv_command ("/.timerimgmirc_kill -m 1 10 /dll -u imgmirc.dll");
		conv_command ("/.timerimgmirc_start 1 1 /dll imgmirc.dll preload");
	}
	conv_command (hCmdConv,"/.timerimgmirc_tryconfigure off");
	SetEvent (hDlgAllowed);
	return 1;
}

int __stdcall keepalive(HWND mWnd, HWND aWnd, char *data, 
						char *parms, BOOL show, BOOL nopause)
{
	verb_conv_echo ("imgmirc: keep_alive received.",2);
	return 1;
}




