/* imgmirc.dll is a plug-in for the mIRC application
* Copyright (C) 2003-2005 Ivan Zvolsky
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version. 
* 
* You should have received a copy of the GNU Lesser General Public
* License along with this library (lesser.txt)
*/
#include "stdafx.h"
#include "imgstorage.h"

#include "convthread.h"
#include "flags.h"

//---------------------------------------------
// Here goes image storage class implementation

_image_item * _img_storage::PutItem (LPCTSTR name)
{
	_image_item * run=head;

	while ((run=run->next) != head)
	{	
		if (strcmp(name,run->imgname)==0) 
			return run;
	}

	run = new _image_item (name);
	run->next=head->next;
	head->next = run;

	if (run->ReadDIB()) return run;
	else
	{
		head->next = run->next;
		delete run;
		return NULL;
	}
}

_img_storage::_img_storage ()
{
	bmpDC = NULL;
	hBmp = NULL;
}

_img_storage::~_img_storage ()
{
	if (hBmp) DeleteObject (hBmp);
	if (bmpDC) DeleteObject (bmpDC);
}

void _img_storage::DrawAll (HDC hDC)
{
	TransparentBlt (hDC,0,0,120,80,bmpDC,0,0,60,40,RGB(255,255,255));
}

// Here goes _image_item class definition
// --------------------------------------

BOOL _image_item::ReadDIB (void)
{
	char				str[1024];
	FILE *				fh;
	BITMAPFILEHEADER	bf;
	WORD				numColors;
	sprintf (str,"%simgmirc\\imgs\\%s",mircDir,imgname);

	// trying to open the file
	fh = fopen (str, "rb");
	if ( fh == NULL ) {return false;}
	// reding the file's header
	if ( fread( &bf, sizeof(bf), 1, fh ) != 1) 
	{ fclose (fh); return false; }
	// make sure it is 'BM'
	if (bf.bfType != 0x4d42)
	{ fclose (fh); return false; }
	// reading bitmap header
	if ( fread( &bi, sizeof(bi), 1, fh ) != 1) 
	{ fclose (fh); return false; }
	// define biClrUsed
	if (!(numColors = (WORD)bi.biClrUsed))
		if ( bi.biBitCount < 16 )
			numColors = 1 << bi.biBitCount;	

	if (bi.biClrUsed == 0)
		bi.biClrUsed = (DWORD)numColors;
	// determining biSizeImage in uncompressed case 
	// ( for example windows 3.x bitmap)
	if ( !bi.biSizeImage )
		bi.biSizeImage = (( ((bi.biWidth * (DWORD)bi.biBitCount) + 31) & ~31) >> 3)
		* bi.biHeight;

	offset = (WORD)bi.biSize + numColors * sizeof(RGBQUAD);

	data.raw_data = new unsigned char [ offset + bi.biSizeImage ];
	fseek(fh, sizeof (bf), SEEK_SET);
	// read header, palette and data
	fread ( data.raw_data, offset + bi.biSizeImage, 1, fh);

	fclose (fh);

	HDC dc = GetDC (NULL);
	memDC = CreateCompatibleDC (dc);
	memBmp = CreateCompatibleBitmap (dc, bi.biWidth, bi.biHeight);
	SelectObject (memDC, (HBITMAP)memBmp);
	ReleaseDC (NULL, dc);

	SetDIBitsToDevice (memDC, 0, 0, bi.biWidth, bi.biHeight,
		0, 0, 0, bi.biHeight, data.raw_data+offset, 
		data.lpbi, DIB_RGB_COLORS);

	return true;
}



_img_storage imagestorage;


