/* imgmirc.dll is a plug-in for the mIRC application
* Copyright (C) 2003-2005 Ivan Zvolsky
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version. 
* 
* You should have received a copy of the GNU Lesser General Public
* License along with this library (lesser.txt)
*/
#include "stdafx.h"
#include "my_funcs.h"
#include "flags.h"
#include "pcre.h"

#include "replstorage.h"
#include "highlights.h"
#include "imgstorage.h"
#include "stdlib.h"

#define __TEXTOUT__ ((BOOL (__stdcall*)(HDC, int, int, UINT, const RECT*, LPCTSTR, UINT, const int*)) func_tags[_TEXTOUT_FUNC_].old_func )
#define __RECV__ ((int (__stdcall*)(SOCKET, char FAR *, int, int)) func_tags[_RECV_FUNC_].old_func )
#define __DRAWTEXT__ ((int (__stdcall*)(HDC, LPCTSTR, int, LPRECT, UINT)) func_tags[_DRAWTEXT_FUNC_].old_func )

struct _func_tag func_tags[NUM_FUNCS_OVERRIDEN];

void return_functions ()
{
	SIZE_T written;
	DWORD op;
	for (int i=0; i<NUM_FUNCS_OVERRIDEN; i++)
		if (func_tags[i].isd)
		{
			VirtualProtect((void*)(func_tags[i].isd),4,PAGE_READWRITE, &op);
			WriteProcessMemory(GetCurrentProcess(), (void*)(func_tags[i].isd),
				(void*)&(func_tags[i].old_func),4,&written);
			VirtualProtect((void*)(func_tags[i].isd),4,op, &op);

			if(written!=4)
			{
				conv_echo ("imgmirc: Unable to rewrite address",4);
				return;
			} else 
				verb_conv_echo (4, "imgmirc: %x address returned", func_tags[i].old_func);
		}
}

void do_intercept (void)
{
	for (int i=0; i<NUM_FUNCS_OVERRIDEN; i++) 
		func_tags[i].isd = NULL;

	if (_PCRE_HIGHLIGHTS_)
	{
		verb_conv_echo ("intercepting recv....",7);
		intercept_function ("WSOCK32.dll", "recv", _RECV_FUNC_, (DWORD)&my_recv);
	}
	if (_PCRE_HIGHLIGHTS_ || _SMILES_)
	{
		verb_conv_echo ("imgmirc: intercepting ExtTextOutA....",7);
		intercept_function ("GDI32.DLL", "ExtTextOutA", _TEXTOUT_FUNC_, 
			_TRANSPARENT_?(DWORD)&my_TranspExtTextOut:(DWORD)&my_SimpleExtTextOut);

		// starting from 6.34 (?) we need to intercept DrawText() - they changed the way text is drawn
		if (mirc_minor_ver >= 34) {
			verb_conv_echo ("intercepting DrawTextA....",7);
			intercept_function ("USER32.dll", "DrawTextA", _DRAWTEXT_FUNC_, 
				_TRANSPARENT_?(DWORD)&my_TransparentDrawText:(DWORD)&my_SimpleDrawText);
		}
	}

}

int intercept_function( LPCTSTR module_name, 
					   LPCTSTR func_name, 
					   int num, 
					   DWORD new_func)
{
	func_tags [num].isd = 0;

	BYTE *pimage = (BYTE*)GetModuleHandle(NULL); 

	IMAGE_DOS_HEADER *idh;
	IMAGE_OPTIONAL_HEADER *ioh;
	IMAGE_SECTION_HEADER *ish;
	IMAGE_IMPORT_DESCRIPTOR *iid;
	DWORD *isd;  //image_thunk_data dword

	idh = (IMAGE_DOS_HEADER*)pimage;
	ioh = (IMAGE_OPTIONAL_HEADER*)(pimage + idh->e_lfanew 
		+ 4 + sizeof(IMAGE_FILE_HEADER));
	ish = (IMAGE_SECTION_HEADER*)((BYTE*)ioh + sizeof(IMAGE_OPTIONAL_HEADER));

	if (idh->e_magic != 0x5A4D) 
	{
		conv_echo ("imgmirc: Not exe hdr",4);
		return 1;
	}

	int i;
	for(i=0; i<16; i++)
		if (strcmp((char*)((ish+ i)->Name) , ".idata") == 0) 
			break;

	if(i==16) 
	{
		verb_conv_echo ("imgmirc: Unable to find .idata section",2);
		iid = (IMAGE_IMPORT_DESCRIPTOR*)(pimage +
			ioh->DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress);
	} else 
		iid = (IMAGE_IMPORT_DESCRIPTOR*)(pimage + (ish +i)->VirtualAddress );

	func_tags[num].old_func = (DWORD)GetProcAddress(GetModuleHandle(module_name), func_name);
	if(func_tags[num].old_func == 0)
	{
		conv_echo (4, "imgmirc: Can`t get address of %s function",func_name);
		return 1;
	} else 
		verb_conv_echo (4,"imgmirc: address of %s is %x", func_name, func_tags[num].old_func);

	while(iid->Name)
	{
		if (_stricmp((char*)(pimage + iid->Name), module_name) ==0 ) break;
		iid++;
	}

	isd = (DWORD*)(pimage + iid->FirstThunk);
	while(*isd!=func_tags[num].old_func && *isd!=0) {
		isd++;
	}


	if(*isd == 0)
	{
		conv_echo ("imgmirc: addr_func not found in .idata", 4);
		return 1;
	}

	DWORD buf = new_func;
	DWORD op;
	SIZE_T written;

	VirtualProtect((void*)(isd),4,PAGE_READWRITE, &op);
	WriteProcessMemory(GetCurrentProcess(), (void*)(isd), (void*)&buf,4,&written);
	VirtualProtect((void*)(isd),4,op, &op);

	if(written!=4)
	{
		conv_echo ("imgmirc: Unable to rewrite address",4);
		return 1;
	}

	func_tags [num].new_func = new_func;
	func_tags [num].isd = isd;

	return 0; //success
}

int WINAPI my_recv(SOCKET s, char FAR *buf, int len, int flags)
{
	int ret;
	ret =  __RECV__ (s, buf, len, flags);

	std::vector<_repl_item *>::iterator r, r_end;
	std::vector<_hl_item *>::iterator i;
	std::vector<_hl_item *>::iterator i_end;

	int gl_pos;
	char buffer[8192]; // FIXME

	if (ret>0) 
	{
		// performing highlighting
		bool flag;
		gl_pos = 0;
		int b_pos = 0;

		while (hl_privmsg->hl_exec (buf, ret, gl_pos) > 0)
		{

			flag = FALSE;

			i_end = hl_list.end();
			for (i=hl_list.begin(); i<i_end && !flag; i++)
				if ((*i)->hl_exec (buf+hl_privmsg->ovector[1], 
					ret-hl_privmsg->ovector[1], 
					0) >0) 
				{
					memcpy (buffer+b_pos, buf+gl_pos, 
						hl_privmsg->ovector[1]-gl_pos+
						(*i)->ovector[1]);
					b_pos += hl_privmsg->ovector[1]-gl_pos+
						(*i)->ovector[1];

					buffer [b_pos++]=' ';
					memcpy (buffer+b_pos, hl_mainstr, 3);
					b_pos += 3;

					flag = TRUE;
					gl_pos = hl_privmsg->ovector[1]+(*i)->ovector[1];
				} 

				if (!flag)
				{
					memcpy (buffer+b_pos, buf+gl_pos, ret-gl_pos);
					b_pos += ret-gl_pos;
					gl_pos = ret;
				}
		}

		if (gl_pos < ret) 
		{
			memcpy (buffer+b_pos, buf+gl_pos, ret-gl_pos);
			b_pos += ret-gl_pos;
			memcpy (buf, buffer, b_pos);
			ret = b_pos;
		}
	}
	return ret;
}

bool my_rect_compare (const RECT & rect1, const RECT & rect2)
{
	return (rect1.top < rect2.top);
}

BOOL WINAPI my_SimpleExtTextOut(
								HDC hdc,          // handle to DC
								int X,            // x-coordinate of reference point
								int Y,            // y-coordinate of reference point
								UINT fuOptions,   // text-output options
								const RECT* lprc, // optional dimensions
								LPCTSTR lpString, // string
								UINT cbCount,     // number of characters in string
								const int* lpDx   // array of spacing values
								)
{
	SIZE sz, sz1;
	std::vector<_repl_item *>::iterator i;
	std::vector<_repl_item *>::iterator i_end;
	std::vector<RECT>::iterator j;
	std::vector<RECT>::iterator j_end;

	int p;
	BOOL ret;
	char * pos;
	HPEN f_pen;

	if (cbCount < 2) return 
		__TEXTOUT__ (hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

	if (_PCRE_HIGHLIGHTS_)
		if (pos=(char*)strstr(lpString, hl_mainstr)) memcpy (pos,"    ",3);

	ret = __TEXTOUT__ (hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

	if (_SMILES_)
	{
		f_pen = (HPEN)SelectObject (hdc, (HPEN)GetStockObject (NULL_PEN));
		i_end = repl_list.end();
		for (i=repl_list.begin(); i<i_end; i++)
		{
			p = 0;
			while ( (*i)->hl->hl_exec ((char *)lpString, cbCount, p) > 0)
			{
				GetTextExtentPoint32 (hdc, lpString + (*i)->hl->ovector[2],
					(*i)->hl->ovector[3] - (*i)->hl->ovector[2],&sz1);
				GetTextExtentPoint32 (hdc, lpString, (*i)->hl->ovector[2], &sz);

				POINT pnt = {X + sz.cx + (sz1.cx - (*i)->rect_image.right)/2, 
					Y - 1 + (sz1.cy - (*i)->rect_image.bottom)/2};
				Rectangle (hdc, X+sz.cx,				Y, X+sz1.cx+sz.cx+1, Y+sz1.cy+1);

				if (sz1.cy > (*i)->rect_image.bottom)
					(*i)->img->JustDrawImage (hdc, &pnt, &(*i)->rect_image); else
					(*i)->img->JustStretchImage (hdc, &pnt, &(*i)->rect_image, sz1.cy);

				p = (*i)->hl->ovector[3];
			}
		}
		SelectObject (hdc, f_pen);
	}
	return ret;
}


BOOL WINAPI my_TranspExtTextOut(
								HDC hdc,          // handle to DC
								int X,            // x-coordinate of reference point
								int Y,            // y-coordinate of reference point
								UINT fuOptions,   // text-output options
								const RECT* lprc, // optional dimensions
								LPCTSTR lpString, // string
								UINT cbCount,     // number of characters in string
								const int* lpDx   // array of spacing values
								)
{
	SIZE sz, sz1;
	std::vector<_repl_item *>::iterator i;
	std::vector<_repl_item *>::iterator i_end;
	std::vector<RECT>::iterator j;
	std::vector<RECT>::iterator j_end;

	int p;
	BOOL ret = true;
	char * pos;
	RECT text_pnt;

	if (cbCount < 2) 
		return __TEXTOUT__ (hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

	if (_PCRE_HIGHLIGHTS_)
		if (pos=(char *)strstr(lpString, hl_mainstr)) memcpy (pos,"    ",3);

	if (_SMILES_)
	{
		i_end = repl_list.end();
		for (i=repl_list.begin(); i<i_end; i++)
		{
			p = 0;

			while ( (*i)->hl->hl_exec ((char *)lpString, cbCount, p) > 0)
			{
				GetTextExtentPoint32 (hdc, lpString + (*i)->hl->ovector[2],
					(*i)->hl->ovector[3] - (*i)->hl->ovector[2],&sz1);
				GetTextExtentPoint32 (hdc, lpString, (*i)->hl->ovector[2], &sz);

				text_pnt.left = X+sz.cx;
				text_pnt.right = X+sz.cx+sz1.cx;
				text_pnt.top = (*i)->hl->ovector[2];
				text_pnt.bottom = (*i)->hl->ovector[3] - (*i)->hl->ovector[2];

				text_list.push_back (text_pnt);

				POINT pnt = {X + sz.cx + (sz1.cx - (*i)->rect_image.right)/2, 
					Y - 1 + (sz1.cy - (*i)->rect_image.bottom)/2};

				text_pnt.top = Y;
				text_pnt.bottom = Y+sz1.cy+1;
				__TEXTOUT__ (hdc, X, Y, fuOptions, &text_pnt, "", 0, lpDx);

				if (sz1.cy > (*i)->rect_image.bottom)
					(*i)->img->TranspDrawImage (hdc, &pnt, &(*i)->rect_image); else
					(*i)->img->TranspStretchImage (hdc, &pnt, &(*i)->rect_image, sz1.cy);

				p = (*i)->hl->ovector[3];
			}
		}

		//sort here
		j_end = text_list.end();
		if (j_end != text_list.begin())
		{
			std::sort (text_list.begin(),text_list.end(),my_rect_compare);
			j_end = text_list.end();

			j = text_list.begin();
			text_pnt.top = Y;
			text_pnt.bottom = Y+sz1.cy+1;
			text_pnt.left = X;
			text_pnt.right = j->left;

			__TEXTOUT__ (hdc, X, Y, fuOptions, &text_pnt, 
				lpString, 
				text_list.begin()->top, lpDx); 

			j_end--;
			for (j=text_list.begin(); j<j_end; j++)
			{
				text_pnt.left = j->right;
				text_pnt.right = (j+1)->left;

				__TEXTOUT__ (hdc, j->right, Y, fuOptions, &text_pnt, 
					lpString+j->top+j->bottom, 
					(j+1)->top-(j->top+j->bottom), lpDx); 
			}

			GetTextExtentPoint32 (hdc, lpString, cbCount, &sz);

			text_pnt.left = j_end->right;
			text_pnt.right = X+sz.cx;

			__TEXTOUT__ (hdc, j_end->right, Y, fuOptions, &text_pnt, 
				lpString+j_end->top+j_end->bottom, 
				cbCount-(j_end->top+j_end->bottom), lpDx); 
		} else 
			ret = __TEXTOUT__ (hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

		text_list.clear();
	} else
		ret = __TEXTOUT__ (hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

	return ret;
}

int WINAPI my_TransparentDrawText(
								  HDC hDC,          // handle to DC
								  LPCTSTR lpchText, // text to draw
								  int nCount,       // text length
								  LPRECT lpRect,    // formatting dimensions
								  UINT uFormat      // text-drawing options
								  )
{
	if (uFormat & DT_CALCRECT 
		|| uFormat == 37 /* dirty hack for multi-tab control like "address book" in 6.34 */)
		return __DRAWTEXT__ (hDC, lpchText, nCount, lpRect, uFormat);
	else {
		my_TranspExtTextOut(hDC, lpRect->left, lpRect->top, 0, lpRect, lpchText, nCount, NULL);
		return lpRect->bottom - lpRect->top;
	}
}

int WINAPI my_SimpleDrawText(
							 HDC hDC,          // handle to DC
							 LPCTSTR lpchText, // text to draw
							 int nCount,       // text length
							 LPRECT lpRect,    // formatting dimensions
							 UINT uFormat      // text-drawing options
							 )
{
	if (uFormat & DT_CALCRECT 
		|| uFormat == 37 /* dirty hack for multi-tab control like "address book" in 6.34 */)
		return __DRAWTEXT__ (hDC, lpchText, nCount, lpRect, uFormat);
	else {
		my_SimpleExtTextOut (hDC, lpRect->left, lpRect->top, 0, lpRect, lpchText, nCount, NULL);
		return lpRect->bottom - lpRect->top;
	}
}

